# coding: utf-8

import datetime
import uuid

from stix2.canonicalization.Canonicalize import canonicalize


class StixCoreRelationship:
    """Main StixCoreRelationship class for OpenCTI

    Manages STIX relationships between entities in the OpenCTI platform.

    :param opencti: instance of :py:class:`~pycti.api.opencti_api_client.OpenCTIApiClient`
    :type opencti: OpenCTIApiClient
    """

    def __init__(self, opencti):
        """Initialize the StixCoreRelationship instance.

        :param opencti: OpenCTI API client instance
        :type opencti: OpenCTIApiClient
        """
        self.opencti = opencti
        self.properties = """
            id
            entity_type
            parent_types
            spec_version
            created_at
            updated_at
            standard_id
            relationship_type
            description
            start_time
            stop_time
            revoked
            confidence
            lang
            created
            modified
            status {
                id
                template {
                  id
                  name
                  color
                }
            }
            createdBy {
                ... on Identity {
                    id
                    standard_id
                    entity_type
                    parent_types
                    spec_version
                    identity_class
                    name
                    description
                    roles
                    contact_information
                    x_opencti_aliases
                    created
                    modified
                    objectLabel {
                        id
                        value
                        color
                    }
                }
                ... on Organization {
                    x_opencti_organization_type
                    x_opencti_reliability
                }
                ... on Individual {
                    x_opencti_firstname
                    x_opencti_lastname
                }
            }
            objectMarking {
                id
                standard_id
                entity_type
                definition_type
                definition
                created
                modified
                x_opencti_order
                x_opencti_color
            }
            objectOrganization {
                id
                standard_id
                name
            }
            objectLabel {
                id
                value
                color
            }
            externalReferences {
                edges {
                    node {
                        id
                        standard_id
                        entity_type
                        source_name
                        description
                        url
                        hash
                        external_id
                        created
                        modified
                        importFiles {
                            edges {
                                node {
                                    id
                                    name
                                    size
                                    metaData {
                                        mimetype
                                        version
                                    }
                                }
                            }
                        }
                    }
                }
            }
            from {
                ... on BasicObject {
                    id
                    entity_type
                    parent_types
                }
                ... on BasicRelationship {
                    id
                    entity_type
                    parent_types
                }
                ... on StixObject {
                    standard_id
                    spec_version
                    created_at
                    updated_at
                }
                ... on AttackPattern {
                    name
                }
                ... on Campaign {
                    name
                }
                ... on CourseOfAction {
                    name
                }
                ... on Individual {
                    name
                }
                ... on Organization {
                    name
                }
                ... on Sector {
                    name
                }
                ... on System {
                    name
                }
                ... on Indicator {
                    name
                }
                ... on Infrastructure {
                    name
                }
                ... on IntrusionSet {
                    name
                }
                ... on Position {
                    name
                }
                ... on City {
                    name
                }
                ... on Country {
                    name
                }
                ... on Region {
                    name
                }
                ... on Malware {
                    name
                }
                ... on ThreatActor {
                    name
                }
                ... on Tool {
                    name
                }
                ... on Vulnerability {
                    name
                }
                ... on Incident {
                    name
                }
                ... on Event {
                    name
                    description
                }
                ... on Channel {
                    name
                    description
                }
                ... on Narrative {
                    name
                    description
                }
                ... on Language {
                    name
                }
                ... on DataComponent {
                    name
                    description
                }
                ... on DataSource {
                    name
                    description
                }
                ... on Case {
                    name
                }
                ... on StixCyberObservable {
                    observable_value
                }
                ... on StixCoreRelationship {
                    standard_id
                    spec_version
                    created_at
                    updated_at
                }
            }
            to {
                ... on BasicObject {
                    id
                    entity_type
                    parent_types
                }
                ... on BasicRelationship {
                    id
                    entity_type
                    parent_types
                }
                ... on StixObject {
                    standard_id
                    spec_version
                    created_at
                    updated_at
                }
                ... on AttackPattern {
                    name
                }
                ... on Campaign {
                    name
                }
                ... on CourseOfAction {
                    name
                }
                ... on Individual {
                    name
                }
                ... on Organization {
                    name
                }
                ... on Sector {
                    name
                }
                ... on System {
                    name
                }
                ... on Indicator {
                    name
                }
                ... on Infrastructure {
                    name
                }
                ... on IntrusionSet {
                    name
                }
                ... on Position {
                    name
                }
                ... on City {
                    name
                }
                ... on Country {
                    name
                }
                ... on Region {
                    name
                }
                ... on Malware {
                    name
                }
                ... on ThreatActor {
                    name
                }
                ... on Tool {
                    name
                }
                ... on Vulnerability {
                    name
                }
                ... on Incident {
                    name
                }
                ... on Event {
                    name
                    description
                }
                ... on Channel {
                    name
                    description
                }
                ... on Narrative {
                    name
                    description
                }
                ... on Language {
                    name
                }
                ... on DataComponent {
                    name
                    description
                }
                ... on DataSource {
                    name
                    description
                }
                ... on Case {
                    name
                }
                ... on StixCyberObservable {
                    observable_value
                }
                ... on StixCoreRelationship {
                    standard_id
                    spec_version
                    created_at
                    updated_at
                }
            }
        """

    @staticmethod
    def generate_id(
        relationship_type, source_ref, target_ref, start_time=None, stop_time=None
    ):
        """Generate a STIX ID for a relationship.

        :param relationship_type: The type of relationship
        :type relationship_type: str
        :param source_ref: The source entity reference ID
        :type source_ref: str
        :param target_ref: The target entity reference ID
        :type target_ref: str
        :param start_time: (optional) The start time of the relationship
        :type start_time: str or datetime.datetime or None
        :param stop_time: (optional) The stop time of the relationship
        :type stop_time: str or datetime.datetime or None
        :return: STIX ID for the relationship
        :rtype: str
        """
        if isinstance(start_time, datetime.datetime):
            start_time = start_time.isoformat()
        if isinstance(stop_time, datetime.datetime):
            stop_time = stop_time.isoformat()

        if start_time is not None and stop_time is not None:
            data = {
                "relationship_type": relationship_type,
                "source_ref": source_ref,
                "target_ref": target_ref,
                "start_time": start_time,
                "stop_time": stop_time,
            }
        elif start_time is not None:
            data = {
                "relationship_type": relationship_type,
                "source_ref": source_ref,
                "target_ref": target_ref,
                "start_time": start_time,
            }
        else:
            data = {
                "relationship_type": relationship_type,
                "source_ref": source_ref,
                "target_ref": target_ref,
            }
        data = canonicalize(data, utf8=False)
        id = str(uuid.uuid5(uuid.UUID("00abedb4-aa42-466c-9c01-fed23315a9b7"), data))
        return "relationship--" + id

    @staticmethod
    def generate_id_from_data(data):
        """Generate a STIX ID from relationship data.

        :param data: Dictionary containing relationship_type, source_ref, target_ref, and optionally start_time/stop_time
        :type data: dict
        :return: STIX ID for the relationship
        :rtype: str
        """
        return StixCoreRelationship.generate_id(
            data["relationship_type"],
            data["source_ref"],
            data["target_ref"],
            data.get("start_time"),
            data.get("stop_time"),
        )

    def list(self, **kwargs):
        """List stix_core_relationship objects.

        :param fromOrToId: the id of an entity (source or target)
        :type fromOrToId: str
        :param elementWithTargetTypes: filter by target types
        :type elementWithTargetTypes: list
        :param fromId: the id of the source entity of the relation
        :type fromId: str
        :param fromTypes: filter by source entity types
        :type fromTypes: list
        :param toId: the id of the target entity of the relation
        :type toId: str
        :param toTypes: filter by target entity types
        :type toTypes: list
        :param relationship_type: the relation type
        :type relationship_type: str
        :param startTimeStart: the start_time date start filter
        :type startTimeStart: str
        :param startTimeStop: the start_time date stop filter
        :type startTimeStop: str
        :param stopTimeStart: the stop_time date start filter
        :type stopTimeStart: str
        :param stopTimeStop: the stop_time date stop filter
        :type stopTimeStop: str
        :param filters: additional filters to apply
        :type filters: dict
        :param first: return the first n rows from the after ID (or the beginning if not set)
        :type first: int
        :param after: ID of the first row for pagination
        :type after: str
        :param orderBy: field to order results by
        :type orderBy: str
        :param orderMode: ordering mode (asc/desc)
        :type orderMode: str
        :param customAttributes: custom attributes to return
        :type customAttributes: str
        :param getAll: whether to retrieve all results
        :type getAll: bool
        :param withPagination: whether to include pagination info
        :type withPagination: bool
        :param search: search keyword
        :type search: str
        :return: List of stix_core_relationship objects
        :rtype: list
        """
        from_or_to_id = kwargs.get("fromOrToId", None)
        element_with_target_types = kwargs.get("elementWithTargetTypes", None)
        from_id = kwargs.get("fromId", None)
        from_types = kwargs.get("fromTypes", None)
        to_id = kwargs.get("toId", None)
        to_types = kwargs.get("toTypes", None)
        relationship_type = kwargs.get("relationship_type", None)
        start_time_start = kwargs.get("startTimeStart", None)
        start_time_stop = kwargs.get("startTimeStop", None)
        stop_time_start = kwargs.get("stopTimeStart", None)
        stop_time_stop = kwargs.get("stopTimeStop", None)
        filters = kwargs.get("filters", None)
        first = kwargs.get("first", 100)
        after = kwargs.get("after", None)
        order_by = kwargs.get("orderBy", None)
        order_mode = kwargs.get("orderMode", None)
        custom_attributes = kwargs.get("customAttributes", None)
        get_all = kwargs.get("getAll", False)
        with_pagination = kwargs.get("withPagination", False)
        search = kwargs.get("search", None)

        self.opencti.app_logger.info(
            "Listing stix_core_relationships",
            {
                "relationship_type": relationship_type,
                "from_or_to_id": from_or_to_id,
                "from_id": from_id,
                "to_id": to_id,
                "element_with_target_types": element_with_target_types,
                "from_types": from_types,
                "to_types": to_types,
                "search": search,
            },
        )
        query = (
            """
                query StixCoreRelationships($fromOrToId: [String], $elementWithTargetTypes: [String], $fromId: [String], $fromTypes: [String], $toId: [String], $toTypes: [String], $relationship_type: [String], $startTimeStart: DateTime, $startTimeStop: DateTime, $stopTimeStart: DateTime, $stopTimeStop: DateTime, $filters: FilterGroup, $first: Int, $after: ID, $orderBy: StixCoreRelationshipsOrdering, $orderMode: OrderingMode, $search: String) {
                    stixCoreRelationships(fromOrToId: $fromOrToId, elementWithTargetTypes: $elementWithTargetTypes, fromId: $fromId, fromTypes: $fromTypes, toId: $toId, toTypes: $toTypes, relationship_type: $relationship_type, startTimeStart: $startTimeStart, startTimeStop: $startTimeStop, stopTimeStart: $stopTimeStart, stopTimeStop: $stopTimeStop, filters: $filters, first: $first, after: $after, orderBy: $orderBy, orderMode: $orderMode, search: $search) {
                        edges {
                            node {
                                """
            + (custom_attributes if custom_attributes is not None else self.properties)
            + """
                        }
                    }
                    pageInfo {
                        startCursor
                        endCursor
                        hasNextPage
                        hasPreviousPage
                        globalCount
                    }
                }
            }
         """
        )
        result = self.opencti.query(
            query,
            {
                "fromOrToId": from_or_to_id,
                "elementWithTargetTypes": element_with_target_types,
                "fromId": from_id,
                "fromTypes": from_types,
                "toId": to_id,
                "toTypes": to_types,
                "relationship_type": relationship_type,
                "startTimeStart": start_time_start,
                "startTimeStop": start_time_stop,
                "stopTimeStart": stop_time_start,
                "stopTimeStop": stop_time_stop,
                "filters": filters,
                "first": first,
                "after": after,
                "orderBy": order_by,
                "orderMode": order_mode,
                "search": search,
            },
        )
        if get_all:
            final_data = []
            data = self.opencti.process_multiple(
                result["data"]["stixCoreRelationships"]
            )
            final_data = final_data + data
            while result["data"]["stixCoreRelationships"]["pageInfo"]["hasNextPage"]:
                after = result["data"]["stixCoreRelationships"]["pageInfo"]["endCursor"]
                self.opencti.app_logger.debug(
                    "Listing StixCoreRelationships", {"after": after}
                )
                result = self.opencti.query(
                    query,
                    {
                        "fromOrToId": from_or_to_id,
                        "elementWithTargetTypes": element_with_target_types,
                        "fromId": from_id,
                        "fromTypes": from_types,
                        "toId": to_id,
                        "toTypes": to_types,
                        "relationship_type": relationship_type,
                        "startTimeStart": start_time_start,
                        "startTimeStop": start_time_stop,
                        "stopTimeStart": stop_time_start,
                        "stopTimeStop": stop_time_stop,
                        "filters": filters,
                        "first": first,
                        "after": after,
                        "orderBy": order_by,
                        "orderMode": order_mode,
                        "search": search,
                    },
                )
                data = self.opencti.process_multiple(
                    result["data"]["stixCoreRelationships"]
                )
                final_data = final_data + data
            return final_data
        else:
            return self.opencti.process_multiple(
                result["data"]["stixCoreRelationships"], with_pagination
            )

    def read(self, **kwargs):
        """Read a stix_core_relationship object.

        :param id: the id of the stix_core_relationship
        :type id: str
        :param fromOrToId: the id of an entity (source or target)
        :type fromOrToId: str
        :param fromId: the id of the source entity of the relation
        :type fromId: str
        :param toId: the id of the target entity of the relation
        :type toId: str
        :param relationship_type: the relation type
        :type relationship_type: str
        :param startTimeStart: the start_time date start filter
        :type startTimeStart: str
        :param startTimeStop: the start_time date stop filter
        :type startTimeStop: str
        :param stopTimeStart: the stop_time date start filter
        :type stopTimeStart: str
        :param stopTimeStop: the stop_time date stop filter
        :type stopTimeStop: str
        :param filters: filters to apply
        :type filters: dict
        :param customAttributes: custom attributes to return
        :type customAttributes: str
        :return: stix_core_relationship object
        :rtype: dict or None
        """
        id = kwargs.get("id", None)
        from_or_to_id = kwargs.get("fromOrToId", None)
        from_id = kwargs.get("fromId", None)
        to_id = kwargs.get("toId", None)
        relationship_type = kwargs.get("relationship_type", None)
        start_time_start = kwargs.get("startTimeStart", None)
        start_time_stop = kwargs.get("startTimeStop", None)
        stop_time_start = kwargs.get("stopTimeStart", None)
        stop_time_stop = kwargs.get("stopTimeStop", None)
        filters = kwargs.get("filters", None)
        custom_attributes = kwargs.get("customAttributes", None)
        if id is not None:
            self.opencti.app_logger.info("Reading stix_core_relationship", {"id": id})
            query = (
                """
                    query StixCoreRelationship($id: String!) {
                        stixCoreRelationship(id: $id) {
                            """
                + (
                    custom_attributes
                    if custom_attributes is not None
                    else self.properties
                )
                + """
                    }
                }
             """
            )
            result = self.opencti.query(query, {"id": id})
            return self.opencti.process_multiple_fields(
                result["data"]["stixCoreRelationship"]
            )
        elif filters is not None:
            result = self.list(filters=filters, customAttributes=custom_attributes)
            if len(result) > 0:
                return result[0]
            else:
                return None
        elif from_id is not None and to_id is not None:
            result = self.list(
                fromOrToId=from_or_to_id,
                fromId=from_id,
                toId=to_id,
                relationship_type=relationship_type,
                startTimeStart=start_time_start,
                startTimeStop=start_time_stop,
                stopTimeStart=stop_time_start,
                stopTimeStop=stop_time_stop,
            )
            if len(result) > 0:
                return result[0]
            else:
                return None
        else:
            self.opencti.app_logger.error("Missing parameters: id or from_id and to_id")
            return None

    def create(self, **kwargs):
        """Create a stix_core_relationship object.

        :param fromId: the id of the source entity
        :type fromId: str
        :param toId: the id of the target entity
        :type toId: str
        :param stix_id: (optional) the STIX ID
        :type stix_id: str
        :param relationship_type: the type of relationship
        :type relationship_type: str
        :param description: (optional) description
        :type description: str
        :param start_time: (optional) start time of the relationship
        :type start_time: str
        :param stop_time: (optional) stop time of the relationship
        :type stop_time: str
        :param revoked: (optional) whether the relationship is revoked
        :type revoked: bool
        :param confidence: (optional) confidence level (0-100)
        :type confidence: int
        :param lang: (optional) language
        :type lang: str
        :param created: (optional) creation date
        :type created: str
        :param modified: (optional) modification date
        :type modified: str
        :param createdBy: (optional) the author ID
        :type createdBy: str
        :param objectMarking: (optional) list of marking definition IDs
        :type objectMarking: list
        :param objectLabel: (optional) list of label IDs
        :type objectLabel: list
        :param externalReferences: (optional) list of external reference IDs
        :type externalReferences: list
        :param killChainPhases: (optional) list of kill chain phase IDs
        :type killChainPhases: list
        :param objectOrganization: (optional) list of organization IDs
        :type objectOrganization: list
        :param x_opencti_workflow_id: (optional) workflow ID
        :type x_opencti_workflow_id: str
        :param x_opencti_stix_ids: (optional) list of additional STIX IDs
        :type x_opencti_stix_ids: list
        :param x_opencti_modified_at: (optional) custom modification date
        :type x_opencti_modified_at: str
        :param coverage_information: (optional) coverage information
        :type coverage_information: list
        :param update: (optional) whether to update if exists (default: False)
        :type update: bool
        :return: stix_core_relationship object
        :rtype: dict or None
        """
        from_id = kwargs.get("fromId", None)
        to_id = kwargs.get("toId", None)
        stix_id = kwargs.get("stix_id", None)
        relationship_type = kwargs.get("relationship_type", None)
        description = kwargs.get("description", None)
        start_time = kwargs.get("start_time", None)
        stop_time = kwargs.get("stop_time", None)
        revoked = kwargs.get("revoked", None)
        confidence = kwargs.get("confidence", None)
        lang = kwargs.get("lang", None)
        created = kwargs.get("created", None)
        modified = kwargs.get("modified", None)
        created_by = kwargs.get("createdBy", None)
        object_marking = kwargs.get("objectMarking", None)
        object_label = kwargs.get("objectLabel", None)
        external_references = kwargs.get("externalReferences", None)
        kill_chain_phases = kwargs.get("killChainPhases", None)
        granted_refs = kwargs.get("objectOrganization", None)
        x_opencti_workflow_id = kwargs.get("x_opencti_workflow_id", None)
        x_opencti_stix_ids = kwargs.get("x_opencti_stix_ids", None)
        x_opencti_modified_at = kwargs.get("x_opencti_modified_at", None)
        coverage_information = kwargs.get("coverage_information", None)
        update = kwargs.get("update", False)
        upsert_operations = kwargs.get("upsert_operations", None)

        self.opencti.app_logger.info(
            "Creating stix_core_relationship",
            {
                "relationship_type": relationship_type,
                "from_id": from_id,
                "to_id": to_id,
            },
        )
        query = """
                mutation StixCoreRelationshipAdd($input: StixCoreRelationshipAddInput!) {
                    stixCoreRelationshipAdd(input: $input) {
                        id
                        standard_id
                        entity_type
                        parent_types
                    }
                }
            """
        result = self.opencti.query(
            query,
            {
                "input": {
                    "fromId": from_id,
                    "toId": to_id,
                    "stix_id": stix_id,
                    "relationship_type": relationship_type,
                    "description": description,
                    "start_time": start_time,
                    "stop_time": stop_time,
                    "revoked": revoked,
                    "confidence": confidence,
                    "lang": lang,
                    "created": created,
                    "modified": modified,
                    "createdBy": created_by,
                    "objectMarking": object_marking,
                    "objectLabel": object_label,
                    "objectOrganization": granted_refs,
                    "externalReferences": external_references,
                    "killChainPhases": kill_chain_phases,
                    "x_opencti_workflow_id": x_opencti_workflow_id,
                    "x_opencti_stix_ids": x_opencti_stix_ids,
                    "x_opencti_modified_at": x_opencti_modified_at,
                    "coverage_information": coverage_information,
                    "update": update,
                    "upsertOperations": upsert_operations,
                }
            },
        )
        return self.opencti.process_multiple_fields(
            result["data"]["stixCoreRelationshipAdd"]
        )

    def update_field(self, **kwargs):
        """Update a stix_core_relationship object field.

        :param id: the stix_core_relationship id
        :type id: str
        :param input: the input of the field
        :type input: list
        :return: The updated stix_core_relationship object
        :rtype: dict or None
        """
        id = kwargs.get("id", None)
        input = kwargs.get("input", None)
        if id is not None and input is not None:
            self.opencti.app_logger.info("Updating stix_core_relationship", {"id": id})
            query = """
                    mutation StixCoreRelationshipEdit($id: ID!, $input: [EditInput]!) {
                        stixCoreRelationshipEdit(id: $id) {
                            fieldPatch(input: $input) {
                                id
                                standard_id
                                entity_type
                            }
                        }
                    }
                """
            result = self.opencti.query(
                query,
                {
                    "id": id,
                    "input": input,
                },
            )
            return self.opencti.process_multiple_fields(
                result["data"]["stixCoreRelationshipEdit"]["fieldPatch"]
            )
        else:
            self.opencti.app_logger.error(
                "[opencti_stix_core_relationship] Missing parameters: id and input",
            )
            return None

    def delete(self, **kwargs):
        """Delete a stix_core_relationship.

        :param id: the stix_core_relationship id
        :type id: str
        :return: None
        """
        id = kwargs.get("id", None)
        if id is not None:
            self.opencti.app_logger.info("Deleting stix_core_relationship", {"id": id})
            query = """
                mutation StixCoreRelationshipEdit($id: ID!) {
                    stixCoreRelationshipEdit(id: $id) {
                        delete
                    }
                }
            """
            self.opencti.query(query, {"id": id})
        else:
            self.opencti.app_logger.error(
                "[opencti_stix_core_relationship] Missing parameters: id"
            )
            return None

    def add_marking_definition(self, **kwargs):
        """Add a Marking-Definition object to stix_core_relationship object (object_marking_refs).

        :param id: the id of the stix_core_relationship
        :type id: str
        :param marking_definition_id: the id of the Marking-Definition
        :type marking_definition_id: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        id = kwargs.get("id", None)
        marking_definition_id = kwargs.get("marking_definition_id", None)
        if id is not None and marking_definition_id is not None:
            custom_attributes = """
                id
                objectMarking {
                    id
                    standard_id
                    entity_type
                    definition_type
                    definition
                    x_opencti_order
                    x_opencti_color
                    created
                    modified
                }
            """
            stix_core_relationship = self.read(
                id=id, customAttributes=custom_attributes
            )
            if stix_core_relationship is None:
                self.opencti.app_logger.error(
                    "Cannot add Marking-Definition, entity not found"
                )
                return False
            if marking_definition_id in stix_core_relationship["objectMarkingIds"]:
                return True
            else:
                self.opencti.app_logger.info(
                    "Adding Marking-Definition to Stix-Domain-Object",
                    {"id": id, "marking_definition_id": marking_definition_id},
                )
                query = """
                   mutation StixCoreRelationshipAddRelation($id: ID!, $input: StixRefRelationshipAddInput!) {
                       stixCoreRelationshipEdit(id: $id) {
                            relationAdd(input: $input) {
                                id
                            }
                       }
                   }
                """
                self.opencti.query(
                    query,
                    {
                        "id": id,
                        "input": {
                            "toId": marking_definition_id,
                            "relationship_type": "object-marking",
                        },
                    },
                )
                return True
        else:
            self.opencti.app_logger.error(
                "Missing parameters: id and marking_definition_id"
            )
            return False

    def remove_marking_definition(self, **kwargs):
        """Remove a Marking-Definition object from stix_core_relationship.

        :param id: the id of the stix_core_relationship
        :type id: str
        :param marking_definition_id: the id of the Marking-Definition
        :type marking_definition_id: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        id = kwargs.get("id", None)
        marking_definition_id = kwargs.get("marking_definition_id", None)
        if id is not None and marking_definition_id is not None:
            self.opencti.app_logger.info(
                "Removing Marking-Definition from stix_core_relationship",
                {"id": id, "marking_definition_id": marking_definition_id},
            )
            query = """
               mutation StixCoreRelationshipRemoveRelation($id: ID!, $toId: StixRef!, $relationship_type: String!) {
                   stixCoreRelationshipEdit(id: $id) {
                        relationDelete(toId: $toId, relationship_type: $relationship_type) {
                            id
                        }
                   }
               }
            """
            self.opencti.query(
                query,
                {
                    "id": id,
                    "toId": marking_definition_id,
                    "relationship_type": "object-marking",
                },
            )
            return True
        else:
            self.opencti.app_logger.error(
                "Missing parameters: id and marking_definition_id"
            )
            return False

    def add_label(self, **kwargs):
        """Add a Label object to stix_core_relationship (labeling).

        :param id: the id of the stix_core_relationship
        :type id: str
        :param label_id: the id of the Label
        :type label_id: str
        :param label_name: (optional) the name of the Label (will create if not exists)
        :type label_name: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        id = kwargs.get("id", None)
        label_id = kwargs.get("label_id", None)
        label_name = kwargs.get("label_name", None)
        if label_name is not None:
            label = self.opencti.label.read(
                filters={
                    "mode": "and",
                    "filters": [{"key": "value", "values": [label_name]}],
                    "filterGroups": [],
                }
            )
            if label:
                label_id = label["id"]
            else:
                label = self.opencti.label.create(value=label_name)
                label_id = label["id"]
        if id is not None and label_id is not None:
            self.opencti.app_logger.info(
                "Adding label to stix-core-relationship",
                {"label_id": label_id, "id": id},
            )
            query = """
               mutation StixCoreRelationshipAddRelation($id: ID!, $input: StixRefRelationshipAddInput!) {
                   stixCoreRelationshipEdit(id: $id) {
                        relationAdd(input: $input) {
                            id
                        }
                   }
               }
            """
            self.opencti.query(
                query,
                {
                    "id": id,
                    "input": {
                        "toId": label_id,
                        "relationship_type": "object-label",
                    },
                },
            )
            return True
        else:
            self.opencti.app_logger.error("Missing parameters: id and label_id")
            return False

    def remove_label(self, **kwargs):
        """Remove a Label object from stix_core_relationship.

        :param id: the id of the stix_core_relationship
        :type id: str
        :param label_id: the id of the Label
        :type label_id: str
        :param label_name: (optional) the name of the Label
        :type label_name: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        id = kwargs.get("id", None)
        label_id = kwargs.get("label_id", None)
        label_name = kwargs.get("label_name", None)
        if label_name is not None:
            label = self.opencti.label.read(
                filters={
                    "mode": "and",
                    "filters": [{"key": "value", "values": [label_name]}],
                    "filterGroups": [],
                }
            )
            if label:
                label_id = label["id"]
        if id is not None and label_id is not None:
            self.opencti.app_logger.info(
                "Removing label from stix_core_relationship",
                {"label_id": label_id, "id": id},
            )
            query = """
               mutation StixCoreRelationshipRemoveRelation($id: ID!, $toId: StixRef!, $relationship_type: String!) {
                   stixCoreRelationshipEdit(id: $id) {
                        relationDelete(toId: $toId, relationship_type: $relationship_type) {
                            id
                        }
                   }
               }
            """
            self.opencti.query(
                query,
                {
                    "id": id,
                    "toId": label_id,
                    "relationship_type": "object-label",
                },
            )
            return True
        else:
            self.opencti.app_logger.error("Missing parameters: id and label_id")
            return False

    def add_external_reference(self, **kwargs):
        """Add an External-Reference object to stix_core_relationship.

        :param id: the id of the stix_core_relationship
        :type id: str
        :param external_reference_id: the id of the External-Reference
        :type external_reference_id: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        id = kwargs.get("id", None)
        external_reference_id = kwargs.get("external_reference_id", None)
        if id is not None and external_reference_id is not None:
            self.opencti.app_logger.info(
                "Adding External-Reference to stix-core-relationship",
                {"external_reference_id": external_reference_id, "id": id},
            )
            query = """
               mutation StixCoreRelationshipEditRelationAdd($id: ID!, $input: StixRefRelationshipAddInput!) {
                   stixCoreRelationshipEdit(id: $id) {
                        relationAdd(input: $input) {
                            id
                        }
                   }
               }
            """
            self.opencti.query(
                query,
                {
                    "id": id,
                    "input": {
                        "toId": external_reference_id,
                        "relationship_type": "external-reference",
                    },
                },
            )
            return True
        else:
            self.opencti.app_logger.error(
                "Missing parameters: id and external_reference_id"
            )
            return False

    def remove_external_reference(self, **kwargs):
        """Remove an External-Reference object from stix_core_relationship.

        :param id: the id of the stix_core_relationship
        :type id: str
        :param external_reference_id: the id of the External-Reference
        :type external_reference_id: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        id = kwargs.get("id", None)
        external_reference_id = kwargs.get("external_reference_id", None)
        if id is not None and external_reference_id is not None:
            self.opencti.app_logger.info(
                "Removing External-Reference from stix_core_relationship",
                {"external_reference_id": external_reference_id, "id": id},
            )
            query = """
               mutation StixCoreRelationshipRemoveRelation($id: ID!, $toId: StixRef!, $relationship_type: String!) {
                   stixCoreRelationshipEdit(id: $id) {
                        relationDelete(toId: $toId, relationship_type: $relationship_type) {
                            id
                        }
                   }
               }
            """
            self.opencti.query(
                query,
                {
                    "id": id,
                    "toId": external_reference_id,
                    "relationship_type": "external-reference",
                },
            )
            return True
        else:
            self.opencti.app_logger.error(
                "Missing parameters: id and external_reference_id"
            )
            return False

    def add_kill_chain_phase(self, **kwargs):
        """Add a Kill-Chain-Phase object to stix_core_relationship object (kill_chain_phases).

        :param id: the id of the stix_core_relationship
        :type id: str
        :param kill_chain_phase_id: the id of the Kill-Chain-Phase
        :type kill_chain_phase_id: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        id = kwargs.get("id", None)
        kill_chain_phase_id = kwargs.get("kill_chain_phase_id", None)
        if id is not None and kill_chain_phase_id is not None:
            self.opencti.app_logger.info(
                "Adding Kill-Chain-Phase to stix-core-relationship",
                {"kill_chain_phase_id": kill_chain_phase_id, "id": id},
            )
            query = """
               mutation StixCoreRelationshipAddRelation($id: ID!, $input: StixRefRelationshipAddInput!) {
                   stixCoreRelationshipEdit(id: $id) {
                        relationAdd(input: $input) {
                            id
                        }
                   }
               }
            """
            self.opencti.query(
                query,
                {
                    "id": id,
                    "input": {
                        "toId": kill_chain_phase_id,
                        "relationship_type": "kill-chain-phase",
                    },
                },
            )
            return True
        else:
            self.opencti.app_logger.error(
                "[opencti_stix_core_relationship] Missing parameters: id and kill_chain_phase_id",
            )
            return False

    def remove_kill_chain_phase(self, **kwargs):
        """Remove a Kill-Chain-Phase object from stix_core_relationship.

        :param id: the id of the stix_core_relationship
        :type id: str
        :param kill_chain_phase_id: the id of the Kill-Chain-Phase
        :type kill_chain_phase_id: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        id = kwargs.get("id", None)
        kill_chain_phase_id = kwargs.get("kill_chain_phase_id", None)
        if id is not None and kill_chain_phase_id is not None:
            self.opencti.app_logger.info(
                "Removing Kill-Chain-Phase from stix_core_relationship",
                {"kill_chain_phase_id": kill_chain_phase_id, "id": id},
            )
            query = """
               mutation StixCoreRelationshipRemoveRelation($id: ID!, $toId: StixRef!, $relationship_type: String!) {
                   stixCoreRelationshipEdit(id: $id) {
                        relationDelete(toId: $toId, relationship_type: $relationship_type) {
                            id
                        }
                   }
               }
            """
            self.opencti.query(
                query,
                {
                    "id": id,
                    "toId": kill_chain_phase_id,
                    "relationship_type": "kill-chain-phase",
                },
            )
            return True
        else:
            self.opencti.app_logger.error(
                "[stix_core_relationship] Missing parameters: id and kill_chain_phase_id"
            )
            return False

    def update_created_by(self, **kwargs):
        """Update the Identity author of a stix_core_relationship (created_by).

        :param id: the id of the stix_core_relationship
        :type id: str
        :param identity_id: the id of the Identity
        :type identity_id: str
        :return: True if successful, False otherwise
        :rtype: bool
        """
        id = kwargs.get("id", None)
        identity_id = kwargs.get("identity_id", None)
        if id is not None:
            self.opencti.app_logger.info(
                "Updating author of stix_core_relationship with Identity",
                {"id": id, "identity_id": identity_id},
            )
            custom_attributes = """
                id
                createdBy {
                    ... on Identity {
                        id
                        standard_id
                        entity_type
                        parent_types
                        name
                        x_opencti_aliases
                        description
                        created
                        modified
                    }
                    ... on Organization {
                        x_opencti_organization_type
                        x_opencti_reliability
                    }
                    ... on Individual {
                        x_opencti_firstname
                        x_opencti_lastname
                    }
                }
            """
            stix_domain_object = self.read(id=id, customAttributes=custom_attributes)
            if stix_domain_object["createdBy"] is not None:
                query = """
                    mutation StixCoreRelationshipEdit($id: ID!, $toId: StixRef! $relationship_type: String!) {
                        stixCoreRelationshipEdit(id: $id) {
                            relationDelete(toId: $toId, relationship_type: $relationship_type) {
                                id
                            }
                        }
                    }
                """
                self.opencti.query(
                    query,
                    {
                        "id": id,
                        "toId": stix_domain_object["createdBy"]["id"],
                        "relationship_type": "created-by",
                    },
                )
            if identity_id is not None:
                # Add the new relation
                query = """
                    mutation StixCoreRelationshipEdit($id: ID!, $input: StixRefRelationshipAddInput!) {
                        stixCoreRelationshipEdit(id: $id) {
                            relationAdd(input: $input) {
                                id
                            }
                        }
                    }
               """
                variables = {
                    "id": id,
                    "input": {
                        "toId": identity_id,
                        "relationship_type": "created-by",
                    },
                }
                self.opencti.query(query, variables)
        else:
            self.opencti.app_logger.error("Missing parameters: id")
            return False

    def import_from_stix2(self, **kwargs):
        """Import a stix_core_relationship from a STIX2 object.

        :param stixRelation: the STIX2 relationship object
        :type stixRelation: dict
        :param extras: extra parameters including created_by_id, object_marking_ids, etc.
        :type extras: dict
        :param update: whether to update if the entity already exists
        :type update: bool
        :param defaultDate: default date to use for start/stop times
        :type defaultDate: str or bool
        :return: stix_core_relationship object
        :rtype: dict or None
        """
        stix_relation = kwargs.get("stixRelation", None)
        extras = kwargs.get("extras", {})
        update = kwargs.get("update", False)
        default_date = kwargs.get("defaultDate", False)
        if stix_relation is not None:
            # Search in extensions
            if "x_opencti_stix_ids" not in stix_relation:
                stix_relation["x_opencti_stix_ids"] = (
                    self.opencti.get_attribute_in_extension("stix_ids", stix_relation)
                )
            if "x_opencti_granted_refs" not in stix_relation:
                stix_relation["x_opencti_granted_refs"] = (
                    self.opencti.get_attribute_in_extension(
                        "granted_refs", stix_relation
                    )
                )
            if "x_opencti_workflow_id" not in stix_relation:
                stix_relation["x_opencti_workflow_id"] = (
                    self.opencti.get_attribute_in_extension(
                        "workflow_id", stix_relation
                    )
                )
            if "x_opencti_modified_at" not in stix_relation:
                stix_relation["x_opencti_modified_at"] = (
                    self.opencti.get_attribute_in_extension(
                        "modified_at", stix_relation
                    )
                )
            if "opencti_upsert_operations" not in stix_relation:
                stix_relation["opencti_upsert_operations"] = (
                    self.opencti.get_attribute_in_extension(
                        "opencti_upsert_operations", stix_relation
                    )
                )

            raw_coverages = (
                stix_relation["coverage"] if "coverage" in stix_relation else []
            )
            coverage_information = list(
                map(
                    lambda cov: {
                        "coverage_name": cov["name"],
                        "coverage_score": cov["score"],
                    },
                    raw_coverages,
                )
            )

            source_ref = stix_relation["source_ref"]
            target_ref = stix_relation["target_ref"]
            return self.create(
                fromId=source_ref,
                toId=target_ref,
                stix_id=stix_relation["id"],
                relationship_type=stix_relation["relationship_type"],
                description=(
                    self.opencti.stix2.convert_markdown(stix_relation["description"])
                    if "description" in stix_relation
                    else None
                ),
                start_time=(
                    stix_relation["start_time"]
                    if "start_time" in stix_relation
                    else default_date
                ),
                stop_time=(
                    stix_relation["stop_time"]
                    if "stop_time" in stix_relation
                    else default_date
                ),
                coverage_information=coverage_information,
                revoked=(
                    stix_relation["revoked"] if "revoked" in stix_relation else None
                ),
                confidence=(
                    stix_relation["confidence"]
                    if "confidence" in stix_relation
                    else None
                ),
                lang=stix_relation["lang"] if "lang" in stix_relation else None,
                created=(
                    stix_relation["created"] if "created" in stix_relation else None
                ),
                modified=(
                    stix_relation["modified"] if "modified" in stix_relation else None
                ),
                createdBy=(
                    extras["created_by_id"] if "created_by_id" in extras else None
                ),
                objectMarking=(
                    extras["object_marking_ids"]
                    if "object_marking_ids" in extras
                    else None
                ),
                objectLabel=(
                    extras["object_label_ids"] if "object_label_ids" in extras else None
                ),
                externalReferences=(
                    extras["external_references_ids"]
                    if "external_references_ids" in extras
                    else None
                ),
                killChainPhases=(
                    extras["kill_chain_phases_ids"]
                    if "kill_chain_phases_ids" in extras
                    else None
                ),
                objectOrganization=(
                    stix_relation["x_opencti_granted_refs"]
                    if "x_opencti_granted_refs" in stix_relation
                    else None
                ),
                x_opencti_workflow_id=(
                    stix_relation["x_opencti_workflow_id"]
                    if "x_opencti_workflow_id" in stix_relation
                    else None
                ),
                x_opencti_stix_ids=(
                    stix_relation["x_opencti_stix_ids"]
                    if "x_opencti_stix_ids" in stix_relation
                    else None
                ),
                x_opencti_modified_at=(
                    stix_relation["x_opencti_modified_at"]
                    if "x_opencti_modified_at" in stix_relation
                    else None
                ),
                update=update,
                upsert_operations=(
                    stix_relation["opencti_upsert_operations"]
                    if "opencti_upsert_operations" in stix_relation
                    else None
                ),
            )
        else:
            self.opencti.app_logger.error(
                "[opencti_stix_core_relationship] Missing parameters: stixObject"
            )
            return None

    def organization_share(self, entity_id, organization_ids, sharing_direct_container):
        """Share element to multiple organizations.

        :param entity_id: the stix_core_relationship id
        :type entity_id: str
        :param organization_ids: the organization IDs to share with
        :type organization_ids: list
        :param sharing_direct_container: whether to share direct container
        :type sharing_direct_container: bool
        :return: None
        """
        query = """
                mutation StixCoreRelationshipEdit($id: ID!, $organizationId: [ID!]!, $directContainerSharing: Boolean) {
                    stixCoreRelationshipEdit(id: $id) {
                        restrictionOrganizationAdd(organizationId: $organizationId, directContainerSharing: $directContainerSharing) {
                          id
                        }
                    }
                }
            """
        self.opencti.query(
            query,
            {
                "id": entity_id,
                "organizationId": organization_ids,
                "directContainerSharing": sharing_direct_container,
            },
        )

    def organization_unshare(
        self, entity_id, organization_ids, sharing_direct_container
    ):
        """Unshare element from multiple organizations.

        :param entity_id: the stix_core_relationship id
        :type entity_id: str
        :param organization_ids: the organization IDs to unshare from
        :type organization_ids: list
        :param sharing_direct_container: whether to unshare direct container
        :type sharing_direct_container: bool
        :return: None
        """
        query = """
                mutation StixCoreRelationshipEdit($id: ID!, $organizationId: [ID!]!, $directContainerSharing: Boolean) {
                    stixCoreRelationshipEdit(id: $id) {
                        restrictionOrganizationDelete(organizationId: $organizationId, directContainerSharing: $directContainerSharing) {
                          id
                        }
                    }
                }
            """
        self.opencti.query(
            query,
            {
                "id": entity_id,
                "organizationId": organization_ids,
                "directContainerSharing": sharing_direct_container,
            },
        )

    def remove_from_draft(self, **kwargs):
        """Remove a stix_core_relationship object from draft (revert).

        :param id: the stix_core_relationship id
        :type id: str
        :return: None
        """
        id = kwargs.get("id", None)
        if id is not None:
            self.opencti.app_logger.info(
                "Draft remove stix_core_relationship", {"id": id}
            )
            query = """
                    mutation StixCoreRelationshipEditDraftRemove($id: ID!) {
                        stixCoreRelationshipEdit(id: $id) {
                            removeFromDraft
                        }
                    }
                """
            self.opencti.query(query, {"id": id})
        else:
            self.opencti.app_logger.error(
                "[stix_core_relationship] Cannot remove from draft, missing parameters: id"
            )
            return None



class OpenCTIApiTrash:
    """OpenCTI Trash API class.

    Manages trash/delete operations.

    :param api: instance of :py:class:`~pycti.api.opencti_api_client.OpenCTIApiClient`
    :type api: OpenCTIApiClient
    """

    def __init__(self, api):
        """Initialize the OpenCTIApiTrash instance.

        :param api: OpenCTI API client instance
        :type api: OpenCTIApiClient
        """
        self.api = api

    def restore(self, operation_id: str):
        """Restore a deleted item from trash.

        :param operation_id: the delete operation id
        :type operation_id: str
        :return: None
        :rtype: None
        """
        query = """
            mutation DeleteOperationRestore($id: ID!) {
                deleteOperationRestore(id: $id)
            }
           """
        self.api.query(
            query,
            {
                "id": operation_id,
            },
        )

    def delete(self, **kwargs):
        """Delete a trash item given its ID.

        :param id: ID for the delete operation on the platform
        :type id: str
        :return: None
        :rtype: None
        """
        delete_operation_id = kwargs.get("id", None)
        if delete_operation_id is None:
            self.api.admin_logger.error(
                "[opencti_trash] Cannot confirm delete, missing parameter: id"
            )
            return None
        query = """
            mutation DeleteOperationConfirm($id: ID!) {
                deleteOperationConfirm(id: $id)
            }
        """
        self.api.query(
            query,
            {
                "id": delete_operation_id,
            },
        )



class OpenCTIApiPublicDashboard:
    """OpenCTI Public Dashboard API class.

    Manages public dashboard operations.

    :param api: instance of :py:class:`~pycti.api.opencti_api_client.OpenCTIApiClient`
    :type api: OpenCTIApiClient
    """

    def __init__(self, api):
        """Initialize the OpenCTIApiPublicDashboard instance.

        :param api: OpenCTI API client instance
        :type api: OpenCTIApiClient
        """
        self.api = api

    def delete(self, **kwargs):
        """Delete a public dashboard.

        :param id: the public dashboard id
        :type id: str
        :return: None
        :rtype: None
        """
        dashboard_id = kwargs.get("id", None)
        if dashboard_id is not None:
            query = """
                mutation PublicDashboardDelete($id: ID!) {
                    publicDashboardDelete(id: $id)
                }
               """
            self.api.query(
                query,
                {
                    "id": dashboard_id,
                },
            )
        else:
            self.api.app_logger.error(
                "[opencti_public_dashboard] Cannot delete public dashboard, missing parameter: id"
            )
            return None



# coding: utf-8

import json
import os
import uuid

import magic
from stix2.canonicalization.Canonicalize import canonicalize


class ExternalReference:
    """Main ExternalReference class for OpenCTI

    Manages external references and citations in the OpenCTI platform.

    :param opencti: instance of :py:class:`~pycti.api.opencti_api_client.OpenCTIApiClient`
    :type opencti: OpenCTIApiClient
    """

    def __init__(self, opencti):
        """Initialize the ExternalReference instance.

        :param opencti: OpenCTI API client instance
        :type opencti: OpenCTIApiClient
        """
        self.opencti = opencti
        self.properties = """
            id
            standard_id
            entity_type
            parent_types
            created_at
            updated_at
            created
            modified
            source_name
            description
            url
            hash
            external_id
        """
        self.properties_with_files = """
            id
            standard_id
            entity_type
            parent_types
            created_at
            updated_at
            created
            modified
            source_name
            description
            url
            hash
            external_id
            importFiles {
                edges {
                    node {
                        id
                        name
                        size
                        metaData {
                            mimetype
                            version
                        }
                    }
                }
            }
        """

    @staticmethod
    def generate_id(url=None, source_name=None, external_id=None):
        """Generate a STIX ID for an External Reference.

        :param url: The URL of the external reference
        :type url: str or None
        :param source_name: The source name
        :type source_name: str or None
        :param external_id: The external ID
        :type external_id: str or None
        :return: STIX ID for the external reference, or None if insufficient parameters
        :rtype: str or None
        """
        if url is not None:
            data = {"url": url}
        elif source_name is not None and external_id is not None:
            data = {"source_name": source_name, "external_id": external_id}
        else:
            return None
        data = canonicalize(data, utf8=False)
        id = str(uuid.uuid5(uuid.UUID("00abedb4-aa42-466c-9c01-fed23315a9b7"), data))
        return "external-reference--" + id

    @staticmethod
    def generate_id_from_data(data):
        """Generate a STIX ID from external reference data.

        :param data: Dictionary containing 'url', 'source_name', or 'external_id' keys
        :type data: dict
        :return: STIX ID for the external reference
        :rtype: str or None
        """
        return ExternalReference.generate_id(
            data.get("url"), data.get("source_name"), data.get("external_id")
        )

    def list(self, **kwargs):
        """List External-Reference objects.

        :param filters: the filters to apply
        :type filters: dict
        :param first: return the first n rows from the after ID (or the beginning if not set)
        :type first: int
        :param after: ID of the first row for pagination
        :type after: str
        :param orderBy: field to order results by
        :type orderBy: str
        :param orderMode: ordering mode (asc/desc)
        :type orderMode: str
        :param customAttributes: custom attributes to return
        :type customAttributes: list
        :param getAll: whether to retrieve all results
        :type getAll: bool
        :param withPagination: whether to include pagination info
        :type withPagination: bool
        :param withFiles: whether to include files
        :type withFiles: bool
        :return: List of External-Reference objects
        :rtype: list
        """
        filters = kwargs.get("filters", None)
        first = kwargs.get("first", 500)
        after = kwargs.get("after", None)
        order_by = kwargs.get("orderBy", None)
        order_mode = kwargs.get("orderMode", None)
        custom_attributes = kwargs.get("customAttributes", None)
        get_all = kwargs.get("getAll", False)
        with_pagination = kwargs.get("withPagination", False)
        with_files = kwargs.get("withFiles", False)

        self.opencti.app_logger.info(
            "Listing External-Reference with filters", {"filters": json.dumps(filters)}
        )
        query = (
            """
            query ExternalReferences($filters: FilterGroup, $first: Int, $after: ID, $orderBy: ExternalReferencesOrdering, $orderMode: OrderingMode) {
                externalReferences(filters: $filters, first: $first, after: $after, orderBy: $orderBy, orderMode: $orderMode) {
                    edges {
                        node {
                            """
            + (
                custom_attributes
                if custom_attributes is not None
                else (self.properties_with_files if with_files else self.properties)
            )
            + """
                        }
                    }
                    pageInfo {
                        startCursor
                        endCursor
                        hasNextPage
                        hasPreviousPage
                        globalCount
                    }
                }
            }
        """
        )
        result = self.opencti.query(
            query,
            {
                "filters": filters,
                "first": first,
                "after": after,
                "orderBy": order_by,
                "orderMode": order_mode,
            },
        )
        if get_all:
            final_data = []
            data = self.opencti.process_multiple(result["data"]["externalReferences"])
            final_data = final_data + data
            while result["data"]["externalReferences"]["pageInfo"]["hasNextPage"]:
                after = result["data"]["externalReferences"]["pageInfo"]["endCursor"]
                self.opencti.app_logger.debug(
                    "Listing External-References", {"after": after}
                )
                result = self.opencti.query(
                    query,
                    {
                        "filters": filters,
                        "first": first,
                        "after": after,
                        "orderBy": order_by,
                        "orderMode": order_mode,
                    },
                )
                data = self.opencti.process_multiple(
                    result["data"]["externalReferences"]
                )
                final_data = final_data + data
            return final_data
        else:
            return self.opencti.process_multiple(
                result["data"]["externalReferences"], with_pagination
            )

    def read(self, **kwargs):
        """Read an External-Reference object.

        :param id: the id of the External-Reference
        :type id: str
        :param filters: the filters to apply if no id provided
        :type filters: dict
        :return: External-Reference object
        :rtype: dict or None
        """
        id = kwargs.get("id", None)
        filters = kwargs.get("filters", None)
        if id is not None:
            self.opencti.app_logger.info("Reading External-Reference", {"id": id})
            query = (
                """
                query ExternalReference($id: String!) {
                    externalReference(id: $id) {
                        """
                + self.properties
                + """
                    }
                }
            """
            )
            result = self.opencti.query(query, {"id": id})
            return self.opencti.process_multiple_fields(
                result["data"]["externalReference"]
            )
        elif filters is not None:
            result = self.list(filters=filters)
            if len(result) > 0:
                return result[0]
            else:
                return None
        else:
            self.opencti.app_logger.error(
                "[opencti_external_reference] Missing parameters: id or filters"
            )
            return None

    def create(self, **kwargs):
        """Create an External Reference object.

        :param stix_id: (optional) the STIX ID
        :type stix_id: str
        :param created: (optional) creation date
        :type created: datetime
        :param modified: (optional) modification date
        :type modified: datetime
        :param source_name: the source name of the External Reference (required if no url)
        :type source_name: str
        :param url: (optional) the URL of the external reference (required if no source_name)
        :type url: str
        :param external_id: (optional) the external ID
        :type external_id: str
        :param description: (optional) description
        :type description: str
        :param x_opencti_stix_ids: (optional) list of additional STIX IDs
        :type x_opencti_stix_ids: list
        :param update: (optional) whether to update if exists (default: False)
        :type update: bool
        :param files: (optional) list of File objects to attach
        :type files: list
        :param filesMarkings: (optional) list of lists of marking definition IDs for each file
        :type filesMarkings: list
        :return: External Reference object
        :rtype: dict or None
        """
        stix_id = kwargs.get("stix_id", None)
        created = kwargs.get("created", None)
        modified = kwargs.get("modified", None)
        source_name = kwargs.get("source_name", None)
        url = kwargs.get("url", None)
        external_id = kwargs.get("external_id", None)
        description = kwargs.get("description", None)
        x_opencti_stix_ids = kwargs.get("x_opencti_stix_ids", None)
        update = kwargs.get("update", False)
        files = kwargs.get("files", None)
        files_markings = kwargs.get("filesMarkings", None)

        if source_name is not None or url is not None:
            self.opencti.app_logger.info(
                "Creating External Reference", {"source_name": source_name}
            )
            query = (
                """
                mutation ExternalReferenceAdd($input: ExternalReferenceAddInput!) {
                    externalReferenceAdd(input: $input) {
                        """
                + self.properties
                + """
                    }
                }
            """
            )
            input_variables = {
                "stix_id": stix_id,
                "created": created,
                "modified": modified,
                "source_name": source_name,
                "external_id": external_id,
                "description": description,
                "url": url,
                "x_opencti_stix_ids": x_opencti_stix_ids,
                "update": update,
                "files": files,
                "filesMarkings": files_markings,
            }
            result = self.opencti.query(query, {"input": input_variables})
            return self.opencti.process_multiple_fields(
                result["data"]["externalReferenceAdd"]
            )
        else:
            self.opencti.app_logger.error(
                "[opencti_external_reference] Missing parameters: source_name and url"
            )
            return None

    def add_file(self, **kwargs):
        """Upload a file in this External-Reference.

        :param id: the External-Reference id
        :type id: str
        :param file_name: the name of the file to upload
        :type file_name: str
        :param data: the file data (if None, reads from file_name path)
        :type data: bytes or None
        :param version: (optional) the file version date
        :type version: datetime
        :param fileMarkings: (optional) list of marking definition IDs for the file
        :type fileMarkings: list
        :param mime_type: (optional) MIME type (default: text/plain)
        :type mime_type: str
        :param no_trigger_import: (optional) don't trigger import (default: False)
        :type no_trigger_import: bool
        :param embedded: (optional) embed the file (default: False)
        :type embedded: bool
        :return: File upload result
        :rtype: dict or None
        """
        id = kwargs.get("id", None)
        file_name = kwargs.get("file_name", None)
        data = kwargs.get("data", None)
        version = kwargs.get("version", None)
        file_markings = kwargs.get("fileMarkings", None)
        mime_type = kwargs.get("mime_type", "text/plain")
        no_trigger_import = kwargs.get("no_trigger_import", False)
        embedded = kwargs.get("embedded", False)
        if id is not None and file_name is not None:
            final_file_name = os.path.basename(file_name)
            query = """
                mutation ExternalReferenceEdit($id: ID!, $file: Upload!, $fileMarkings: [String], $version: DateTime, $noTriggerImport: Boolean, $embedded: Boolean) {
                    externalReferenceEdit(id: $id) {
                        importPush(file: $file, fileMarkings: $fileMarkings, version: $version, noTriggerImport: $noTriggerImport, embedded: $embedded) {
                            id
                            name
                        }
                    }
                }
             """
            if data is None:
                data = open(file_name, "rb")
                if file_name.endswith(".json"):
                    mime_type = "application/json"
                else:
                    mime_type = magic.from_file(file_name, mime=True)
            self.opencti.app_logger.info(
                "Uploading a file in External-Reference",
                {"file": final_file_name, "id": id},
            )
            return self.opencti.query(
                query,
                {
                    "id": id,
                    "file": (self.opencti.file(final_file_name, data, mime_type)),
                    "fileMarkings": file_markings,
                    "version": version,
                    "noTriggerImport": (
                        no_trigger_import
                        if isinstance(no_trigger_import, bool)
                        else no_trigger_import == "True"
                    ),
                    "embedded": embedded,
                },
            )
        else:
            self.opencti.app_logger.error(
                "[opencti_external_reference] Missing parameters: id or file_name"
            )
            return None

    def update_field(self, **kwargs):
        """Update an External Reference object field.

        :param id: the External Reference id
        :type id: str
        :param input: the input of the field
        :type input: list
        :return: The updated External Reference object
        :rtype: dict or None
        """
        id = kwargs.get("id", None)
        input = kwargs.get("input", None)
        if id is not None and input is not None:
            self.opencti.app_logger.info("Updating External-Reference", {"id": id})
            query = """
                    mutation ExternalReferenceEdit($id: ID!, $input: [EditInput]!) {
                        externalReferenceEdit(id: $id) {
                            fieldPatch(input: $input) {
                                id
                            }
                        }
                    }
                """
            result = self.opencti.query(query, {"id": id, "input": input})
            return self.opencti.process_multiple_fields(
                result["data"]["externalReferenceEdit"]["fieldPatch"]
            )
        else:
            self.opencti.app_logger.error(
                "[opencti_external_reference] Missing parameters: id and key and value"
            )
            return None

    def delete(self, id):
        """Delete an External-Reference object.

        :param id: the id of the External-Reference to delete
        :type id: str
        :return: None
        """
        self.opencti.app_logger.info("Deleting External-Reference", {"id": id})
        query = """
             mutation ExternalReferenceEdit($id: ID!) {
                 externalReferenceEdit(id: $id) {
                     delete
                 }
             }
         """
        self.opencti.query(query, {"id": id})

    def list_files(self, **kwargs):
        """List files attached to an External-Reference.

        :param id: the id of the External-Reference
        :type id: str
        :return: List of files
        :rtype: list
        """
        id = kwargs.get("id", None)
        self.opencti.app_logger.debug("Listing files of External-Reference", {"id": id})
        query = """
            query externalReference($id: String!) {
                externalReference(id: $id) {
                    importFiles {
                        edges {
                            node {
                                id
                                name
                                size
                                metaData {
                                    mimetype
                                    version
                                }
                            }
                        }
                    }
                }
            }
        """
        result = self.opencti.query(query, {"id": id})
        entity = self.opencti.process_multiple_fields(
            result["data"]["externalReference"]
        )
        return entity["importFiles"]



class OpenCTIApiNotification:
    """OpenCTI Notification API class.

    Manages notification operations.

    :param api: instance of :py:class:`~pycti.api.opencti_api_client.OpenCTIApiClient`
    :type api: OpenCTIApiClient
    """

    def __init__(self, api):
        """Initialize the OpenCTIApiNotification instance.

        :param api: OpenCTI API client instance
        :type api: OpenCTIApiClient
        """
        self.api = api

    def delete(self, **kwargs):
        """Delete a notification.

        :param id: the notification id
        :type id: str
        :return: None
        :rtype: None
        """
        notification_id = kwargs.get("id", None)
        self.api.app_logger.info(
            "Deleting notification", {"notification_id": notification_id}
        )
        query = """
            mutation notificationDelete($id: ID!) {
                notificationDelete(id: $id)
            }
           """
        self.api.query(query, {"id": notification_id})

    def update_field(self, **kwargs):
        """Update a notification field.

        :param id: the notification id
        :type id: str
        :param input: the input fields to update (list of key/value dicts)
        :type input: list
        :return: None
        :rtype: None
        """
        notification_id = kwargs.get("id", None)
        field_input = kwargs.get("input", None)
        if field_input is None:
            return None
        for input_value in field_input:
            if input_value["key"] == "is_read":
                is_read_value = bool(input_value["value"][0])
                self.mark_as_read(notification_id, is_read_value)

    def mark_as_read(self, notification_id: str, read: bool):
        """Mark a notification as read or unread.

        :param notification_id: the notification id
        :type notification_id: str
        :param read: whether to mark as read (True) or unread (False)
        :type read: bool
        :return: None
        :rtype: None
        """
        self.api.app_logger.info(
            "Marking notification as read",
            {"notification_id": notification_id, "read": read},
        )
        query = """
                mutation notificationMarkRead($id: ID!, $read: Boolean!) {
                    notificationMarkRead(id: $id, read: $read) {
                        id
                    }
                }
               """
        self.api.query(query, {"id": notification_id, "read": read})



"""These are the custom STIX properties and observation types used internally by OpenCTI."""

from enum import Enum

from stix2 import CustomObject, CustomObservable, ExternalReference
from stix2.properties import (
    ListProperty,
    ReferenceProperty,
    StringProperty,
    TimestampProperty,
)
from stix2.utils import NOW


class CaseInsensitiveEnum(Enum):
    """Base Enum class with case-insensitive value lookup."""

    @classmethod
    def has_value(cls, value: str) -> bool:
        """Check if the enum contains the given value (case-insensitive).

        :param value: Value to check
        :type value: str
        :return: True if value exists in enum, False otherwise
        :rtype: bool
        """
        lower_values = [v.lower() for v in cls._value2member_map_]
        return value.lower() in lower_values


class StixCyberObservableTypes(CaseInsensitiveEnum):
    """Enumeration of STIX Cyber Observable types supported by OpenCTI."""

    AUTONOMOUS_SYSTEM = "Autonomous-System"
    DIRECTORY = "Directory"
    DOMAIN_NAME = "Domain-Name"
    EMAIL_ADDR = "Email-Addr"
    EMAIL_MESSAGE = "Email-Message"
    EMAIL_MIME_PART_TYPE = "Email-Mime-Part-Type"
    ARTIFACT = "Artifact"
    FILE = "File"
    X509_CERTIFICATE = "X509-Certificate"
    IPV4_ADDR = "IPv4-Addr"
    IPV6_ADDR = "IPv6-Addr"
    MAC_ADDR = "Mac-Addr"
    MUTEX = "Mutex"
    NETWORK_TRAFFIC = "Network-Traffic"
    PROCESS = "Process"
    SOFTWARE = "Software"
    URL = "Url"
    USER_ACCOUNT = "User-Account"
    WINDOWS_REGISTRY_KEY = "Windows-Registry-Key"
    WINDOWS_REGISTRY_VALUE_TYPE = "Windows-Registry-Value-Type"
    HOSTNAME = "Hostname"
    CRYPTOGRAPHIC_KEY = "Cryptographic-Key"
    CRYPTOCURRENCY_WALLET = "Cryptocurrency-Wallet"
    TEXT = "Text"
    USER_AGENT = "User-Agent"
    BANK_ACCOUNT = "Bank-Account"
    PHONE_NUMBER = "Phone-Number"
    CREDENTIAL = "Credential"
    TRACKING_NUMBER = "Tracking-Number"
    PAYMENT_CARD = "Payment-Card"
    MEDIA_CONTENT = "Media-Content"
    SIMPLE_OBSERVABLE = "Simple-Observable"
    PERSONA = "Persona"
    SSH_KEY = "SSH-Key"


class IdentityTypes(CaseInsensitiveEnum):
    """Enumeration of Identity types supported by OpenCTI."""

    SECTOR = "Sector"
    ORGANIZATION = "Organization"
    INDIVIDUAL = "Individual"
    SYSTEM = "System"
    SECURITYPLATFORM = "SecurityPlatform"


class ThreatActorTypes(CaseInsensitiveEnum):
    """Enumeration of Threat Actor types supported by OpenCTI."""

    THREAT_ACTOR_GROUP = "Threat-Actor-Group"
    THREAT_ACTOR_INDIVIDUAL = "Threat-Actor-Individual"


class LocationTypes(CaseInsensitiveEnum):
    """Enumeration of Location types supported by OpenCTI."""

    REGION = "Region"
    COUNTRY = "Country"
    ADMINISTRATIVE_AREA = "Administrative-Area"
    CITY = "City"
    POSITION = "Position"


class ContainerTypes(CaseInsensitiveEnum):
    """Enumeration of Container types supported by OpenCTI."""

    NOTE = "Note"
    OBSERVED_DATA = "Observed-Data"
    OPINION = "Opinion"
    REPORT = "Report"
    GROUPING = "Grouping"
    CASE = "Case"


class StixMetaTypes(CaseInsensitiveEnum):
    """Enumeration of STIX Meta Object types supported by OpenCTI."""

    MARKING_DEFINITION = "Marking-Definition"
    LABEL = "Label"
    EXTERNAL_REFERENCE = "External-Reference"
    KILL_CHAIN_PHASE = "Kill-Chain-Phase"


class MultipleRefRelationship(CaseInsensitiveEnum):
    """Enumeration of relationship types that can have multiple references."""

    OPERATING_SYSTEM = "operating-system"
    SAMPLE = "sample"
    CONTAINS = "contains"
    RESOLVES_TO = "obs_resolves-to"
    BELONGS_TO = "obs_belongs-to"
    TO = "to"
    CC = "cc"
    BCC = "bcc"
    ENCAPSULATES = "encapsulates"
    OPENED_CONNECTION = "opened-connection"
    CHILD = "child"
    BODY_MULTIPART = "body-multipart"
    VALUES = "values"
    SERVICE_DLL = "service-dll"
    INSTALLED_SOFTWARE = "installed-software"
    RELATION_ANALYSIS_SCO = "analysis-sco"


# Custom objects


@CustomObject(
    "case-incident",
    [
        ("name", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        ("description", StringProperty()),
        ("severity", StringProperty()),
        ("priority", StringProperty()),
        ("response_types", ListProperty(StringProperty)),
        ("x_opencti_workflow_id", StringProperty()),
        ("x_opencti_assignee_ids", ListProperty(StringProperty)),
        ("external_references", ListProperty(ExternalReference)),
        (
            "object_refs",
            ListProperty(
                ReferenceProperty(valid_types=["SCO", "SDO", "SRO"], spec_version="2.1")
            ),
        ),
    ],
)
class CustomObjectCaseIncident:
    """Custom STIX2 Case-Incident object for OpenCTI.

    Represents a case-incident container with associated metadata including
    name, description, severity, priority, and response types.

    :param name: Name of the case incident (required)
    :type name: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param description: Description of the case incident
    :type description: str
    :param severity: Severity level of the incident
    :type severity: str
    :param priority: Priority level of the incident
    :type priority: str
    :param response_types: List of response types
    :type response_types: list
    :param x_opencti_workflow_id: OpenCTI workflow identifier
    :type x_opencti_workflow_id: str
    :param x_opencti_assignee_ids: List of assignee identifiers
    :type x_opencti_assignee_ids: list
    :param external_references: List of external references
    :type external_references: list
    :param object_refs: List of referenced STIX objects
    :type object_refs: list
    """

    pass


@CustomObject(
    "case-rfi",
    [
        ("name", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        ("description", StringProperty()),
        ("severity", StringProperty()),
        ("priority", StringProperty()),
        ("information_types", ListProperty(StringProperty)),
        ("x_opencti_workflow_id", StringProperty()),
        ("x_opencti_assignee_ids", ListProperty(StringProperty)),
        ("external_references", ListProperty(ExternalReference)),
        (
            "object_refs",
            ListProperty(
                ReferenceProperty(valid_types=["SCO", "SDO", "SRO"], spec_version="2.1")
            ),
        ),
    ],
)
class CustomObjectCaseRfi:
    """Custom STIX2 Case-RFI (Request For Information) object for OpenCTI.

    Represents a request for information container with associated metadata
    including name, description, severity, priority, and information types.

    :param name: Name of the RFI case (required)
    :type name: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param description: Description of the RFI case
    :type description: str
    :param severity: Severity level of the RFI
    :type severity: str
    :param priority: Priority level of the RFI
    :type priority: str
    :param information_types: List of information types requested
    :type information_types: list
    :param x_opencti_workflow_id: OpenCTI workflow identifier
    :type x_opencti_workflow_id: str
    :param x_opencti_assignee_ids: List of assignee identifiers
    :type x_opencti_assignee_ids: list
    :param external_references: List of external references
    :type external_references: list
    :param object_refs: List of referenced STIX objects
    :type object_refs: list
    """

    pass


@CustomObject(
    "task",
    [
        ("name", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        ("description", StringProperty()),
        (
            "due_date",
            TimestampProperty(
                default=lambda: NOW, precision="millisecond", precision_constraint="min"
            ),
        ),
        ("x_opencti_workflow_id", StringProperty()),
        ("x_opencti_assignee_ids", ListProperty(StringProperty)),
        (
            "object_refs",
            ListProperty(
                ReferenceProperty(valid_types=["SCO", "SDO", "SRO"], spec_version="2.1")
            ),
        ),
    ],
)
class CustomObjectTask:
    """Custom STIX2 Task object for OpenCTI.

    Represents a task with associated metadata including name, description,
    due date, and assignees.

    :param name: Name of the task (required)
    :type name: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param description: Description of the task
    :type description: str
    :param due_date: Due date timestamp for the task
    :type due_date: datetime
    :param x_opencti_workflow_id: OpenCTI workflow identifier
    :type x_opencti_workflow_id: str
    :param x_opencti_assignee_ids: List of assignee identifiers
    :type x_opencti_assignee_ids: list
    :param object_refs: List of referenced STIX objects
    :type object_refs: list
    """

    pass


@CustomObject(
    "channel",
    [
        ("name", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        ("description", StringProperty()),
        ("aliases", ListProperty(StringProperty)),
        ("channel_types", ListProperty(StringProperty)),
        ("x_opencti_workflow_id", StringProperty()),
        ("x_opencti_assignee_ids", ListProperty(StringProperty)),
        ("external_references", ListProperty(ExternalReference)),
    ],
)
class CustomObjectChannel:
    """Custom STIX2 Channel object for OpenCTI.

    Represents a communication channel with associated metadata including
    name, description, aliases, and channel types.

    :param name: Name of the channel (required)
    :type name: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param description: Description of the channel
    :type description: str
    :param aliases: List of alternative names for the channel
    :type aliases: list
    :param channel_types: List of channel types
    :type channel_types: list
    :param x_opencti_workflow_id: OpenCTI workflow identifier
    :type x_opencti_workflow_id: str
    :param x_opencti_assignee_ids: List of assignee identifiers
    :type x_opencti_assignee_ids: list
    :param external_references: List of external references
    :type external_references: list
    """

    pass


# Custom observables


@CustomObservable(
    "hostname",
    [
        ("value", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["value"],
)
class CustomObservableHostname:
    """Custom STIX2 Hostname observable for OpenCTI.

    Represents a hostname cyber observable with its associated value.

    :param value: The hostname value (required)
    :type value: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "text",
    [
        ("value", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["value"],
)
class CustomObservableText:
    """Custom STIX2 Text observable for OpenCTI.

    Represents a generic text cyber observable with its associated value.

    :param value: The text value (required)
    :type value: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "payment-card",
    [
        ("value", StringProperty(required=True)),
        ("card_number", StringProperty(required=True)),
        ("expiration_date", StringProperty(required=False)),
        ("cvv", StringProperty(required=False)),
        ("holder_name", StringProperty(required=False)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["card_number"],
)
class CustomObservablePaymentCard:
    """Custom STIX2 Payment Card observable for OpenCTI.

    Represents a payment card cyber observable with card details.

    :param value: Display value for the payment card (required)
    :type value: str
    :param card_number: The payment card number (required)
    :type card_number: str
    :param expiration_date: Card expiration date
    :type expiration_date: str
    :param cvv: Card verification value
    :type cvv: str
    :param holder_name: Name of the card holder
    :type holder_name: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "bank-account",
    [
        ("value", StringProperty(required=True)),
        ("iban", StringProperty(required=True)),
        ("bic", StringProperty(required=False)),
        ("account_number", StringProperty(required=False)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["iban"],
)
class CustomObservableBankAccount:
    """Custom STIX2 Bank Account observable for OpenCTI.

    Represents a bank account cyber observable with account details.

    :param value: Display value for the bank account (required)
    :type value: str
    :param iban: International Bank Account Number (required)
    :type iban: str
    :param bic: Bank Identifier Code
    :type bic: str
    :param account_number: Bank account number
    :type account_number: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "credential",
    [
        ("value", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["value"],
)
class CustomObservableCredential:
    """Custom STIX2 Credential observable for OpenCTI.

    Represents a credential cyber observable such as a password or access token.

    :param value: The credential value (required)
    :type value: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "cryptocurrency-wallet",
    [
        ("value", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["value"],
)
class CustomObservableCryptocurrencyWallet:
    """Custom STIX2 Cryptocurrency Wallet observable for OpenCTI.

    Represents a cryptocurrency wallet address cyber observable.

    :param value: The wallet address value (required)
    :type value: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "phone-number",
    [
        ("value", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["value"],
)
class CustomObservablePhoneNumber:
    """Custom STIX2 Phone Number observable for OpenCTI.

    Represents a phone number cyber observable.

    :param value: The phone number value (required)
    :type value: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "tracking-number",
    [
        ("value", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["value"],
)
class CustomObservableTrackingNumber:
    """Custom STIX2 Tracking Number observable for OpenCTI.

    Represents a tracking number cyber observable (e.g., package tracking).

    :param value: The tracking number value (required)
    :type value: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "user-agent",
    [
        ("value", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["value"],
)
class CustomObservableUserAgent:
    """Custom STIX2 User-Agent observable for OpenCTI.

    Represents a User-Agent string cyber observable from HTTP headers.

    :param value: The User-Agent string value (required)
    :type value: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "media-content",
    [
        ("title", StringProperty()),
        ("description", StringProperty()),
        ("content", StringProperty()),
        ("media_category", StringProperty()),
        ("url", StringProperty(required=True)),
        ("publication_date", TimestampProperty()),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["url"],
)
class CustomObservableMediaContent:
    """Custom STIX2 Media-Content observable for OpenCTI.

    Represents a media content cyber observable such as articles or posts.

    :param title: Title of the media content
    :type title: str
    :param description: Description of the media content
    :type description: str
    :param content: The actual content body
    :type content: str
    :param media_category: Category of the media
    :type media_category: str
    :param url: URL of the media content (required)
    :type url: str
    :param publication_date: Publication date timestamp
    :type publication_date: datetime
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "persona",
    [
        ("persona_name", StringProperty(required=True)),
        ("persona_type", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["persona_name", "persona_type"],
)
class CustomObservablePersona:
    """Custom STIX2 Persona observable for OpenCTI.

    Represents a persona or online identity cyber observable.

    :param persona_name: Name of the persona (required)
    :type persona_name: str
    :param persona_type: Type of the persona (required)
    :type persona_type: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "cryptographic-key",
    [
        ("value", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["value"],
)
class CustomObservableCryptographicKey:
    """Custom STIX2 Cryptographic-Key observable for OpenCTI.

    Represents a cryptographic key cyber observable such as API keys or encryption keys.

    :param value: The cryptographic key value (required)
    :type value: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass


@CustomObservable(
    "ssh-key",
    [
        ("value", StringProperty(required=True)),
        ("spec_version", StringProperty(fixed="2.1")),
        (
            "object_marking_refs",
            ListProperty(
                ReferenceProperty(valid_types="marking-definition", spec_version="2.1")
            ),
        ),
    ],
    ["value"],
)
class CustomObservableSshKey:
    """Custom STIX2 SSH-Key observable for OpenCTI.

    Represents an SSH key cyber observable such as public or private SSH keys.

    :param value: The SSH key value (required)
    :type value: str
    :param spec_version: STIX specification version, fixed to "2.1"
    :type spec_version: str
    :param object_marking_refs: List of marking definition references
    :type object_marking_refs: list
    """

    pass
